unit uSynchronizeDirectories;
{$INCLUDE 'DirSync.inc'}

interface

uses
  SysUtils, Windows, Classes,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  uSystem, uFiles, uFAR, uLog,
  uDirSyncConsts, uMessages, uSynchronizeDirectoriesDialog, uConfirmations,
  uOptions, uDifferences;

type
  TSynchronizeTask = class;

  TSynchronizeTask = class(TFarEditorManaged)
  private
    fLeftDirectory: string;
    fRightDirectory: string;
    fProgress: TFarProgress;
    fCopyFileSourceFileName: string;
    fCopyFileDestinationFileName: string;
    fCopyFileAborted: boolean;
    fSilentMode: boolean;
  protected
    type
      TSyncFileActionFlag = (afAsk, afYes, afNo, afAll, afNone);
      TSyncFileDirection = (sfdSkip, sfdSkipAll, sfdCopyToLeft, sfdCopyToLeftAll, sfdCopyToRight, sfdCopyToRightAll);
      TSyncFileRetrySkipCancel = (rscRetry, rscSkip, rscCancel);
  protected
    function RemoveNTSchematicsFromFileName(const FileName: string): string;
    function GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
    function LoadFiles(out List: TDifferenceList; const FileName: string): boolean;
    procedure ShowProgress(const SourceFileName, DestinationFileName: string; Progress, Max: int64; Msg1, Msg2: TMessages);
    procedure ShowFileCopyProgress(Progress, Max: int64);
    function DeleteDirectoryMissingOnTheOtherPanel(const FileName: string; var CanDeleteDirectories, CanDeleteFiles, CanDeleteReadOnlyDirectories, CanDeleteReadOnlyFiles: TSyncFileActionFlag; var SkipAllFailedDeletes: boolean; DeleteOnLeft: boolean): boolean;
    function DeleteFileMissingOnTheOtherPanel(const FileName: string; var CanDelete, CanDeleteReadOnly: TSyncFileActionFlag; var SkipAllFailedDeletes: boolean; DeleteOnLeft: boolean): boolean;
    function AskOverwriteConfirmation(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; const FromRightToLeft: boolean): boolean;
    function AskDeleteConfirmation(const FileName: string; var CanDelete, CanDeleteReadOnly: TSyncFileActionFlag; const DeleteOnLeft: boolean): boolean;
    function AskWhatToDoWithDifferentFiles(const LeftDirectory, RightDirectory, FileName: string): TSyncFileDirection;
    function AskRetrySkipCancel1(Msg: TMessages; const FileName: string; var SkipAll: boolean): TSyncFileRetrySkipCancel;
    function AskRetrySkipCancel2(Msg1, Msg2: TMessages; const FileName1, FileName2: string; var SkipAll: boolean): TSyncFileRetrySkipCancel;
    function InternalCopyFileWithProgress(const SourceFileName, DestinationFileName: string; out ErrorCode: DWORD): boolean;
    function CopyFile(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
    function CopyDirectory(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
    function SyncFile(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
    procedure EscapePressedEvent(Sender: TObject; var Value: boolean);
    procedure SetupKeyBar;
    procedure BeforeClose(var CanClose: boolean; var Result: TIntPtr); override;
    procedure AfterShowEditor; override;
  public
    constructor Create(const LeftDirectory, RightDirectory: string);
    destructor Destroy; override;
    function ShowEditor: boolean;
    function Execute(out ReturnToEditor: boolean): boolean;
    procedure VisualCompare(LeftFileName, RightFileName: string);
    procedure Edit(const FileName: string);
    property LeftDirectory: string read fLeftDirectory;
    property RightDirectory: string read fRightDirectory;
    property SilentMode: boolean read fSilentMode write fSilentMode;
  end;

implementation

uses
  uDirSync;

{ TSyncFileInfo }

constructor TSynchronizeTask.Create(const LeftDirectory, RightDirectory: string);
var
  Options: TDirSyncOptions;
  TextStream: TTextStream;
begin
  inherited Create(TFarUtils.TempFileName);
  fLeftDirectory := IncludeTrailingPathDelimiter(LeftDirectory);
  fRightDirectory := IncludeTrailingPathDelimiter(RightDirectory);
  fProgress := nil;
  Options := TDirSyncOptions.Create;
  try
    Options.Load;
    fSilentMode := Options.SilentModeByDefault;
    TextStream := TTextStream.Create(Self.FileName, fmCreate);
    try
      try
        TextStream.Encoding := {$IFDEF UNICODE} feUTF8 {$ELSE} feAnsi {$ENDIF} ;
        TextStream.WriteLine(Format(GetMsgPas(MReportFileTitle), [LeftDirectory, RightDirectory]));
        if Options.ShowDescriptionIsSyncFile then
          TextStream.WriteLine(Format('%s%s', [GetMsgPas(MReportFileNotes1), GetMsgPas(MReportFileNotes2)]));
        TextStream.WriteLine('');
      finally
        FreeAndNil(TextStream);
      end;
    except
      if FileExists(Self.FileName) then
        SysUtils.DeleteFile(Self.FileName);
      Raise;
    end;
  finally
    FreeAndNil(Options);
    end;
end;

function TSynchronizeTask.DeleteDirectoryMissingOnTheOtherPanel(const FileName: string; var CanDeleteDirectories, CanDeleteFiles, CanDeleteReadOnlyDirectories, CanDeleteReadOnlyFiles: TSyncFileActionFlag; var SkipAllFailedDeletes: boolean; DeleteOnLeft: boolean): boolean;
var
  Attr: DWORD;
  SR: TSearchRec;
  Dir: string;
  ContentDeleted: boolean;
begin
  Result := False;
  ShowProgress(FileName, '', 0, 1, MDeletingDirectory, MNoLngStringDefined);
  //Log('Deleting directory "%s"', [FileName]);
  repeat
    if not GetFileAttributes(FileName, Attr) then begin
      //Log('- File does not exist');
      Result := True;
      Break;
    end
    else begin
      if not AskDeleteConfirmation(FileName, CanDeleteDirectories, CanDeleteReadOnlyDirectories, DeleteOnLeft) then begin
        //Log('- User did not allow deleting the destination file');
        Break;
      end;
      // The user allowed me to delete the directory. First delete the content.
      Dir := IncludeTrailingPathDelimiter(FileName);
      //Log('Searching "%s*.*"', [Dir]);
      ContentDeleted := True;
      if FindFirst(Dir + '*.*', faAnyFile, SR) = 0 then
        try
          repeat
            //Log('Found item: "%s" (directory=%d)', [SR.Name, Integer((SR.Attr and faDirectory) <> 0)]);
            if (SR.Attr and faDirectory) <> 0 then begin
              if (SR.Name <> '.') and (SR.Name <> '..') then begin
                if not DeleteDirectoryMissingOnTheOtherPanel(Dir + SR.Name, CanDeleteDirectories, CanDeleteFiles, CanDeleteReadOnlyDirectories, CanDeleteReadOnlyFiles, SkipAllFailedDeletes, DeleteOnLeft) then begin
                  ContentDeleted := False;
                  //Log('Failed to delete');
                end;
              end;
            end
            else begin
              if not DeleteFileMissingOnTheOtherPanel(Dir + SR.Name, CanDeleteFiles, CanDeleteReadOnlyFiles, SkipAllFailedDeletes, DeleteOnLeft) then begin
                ContentDeleted := False;
                //Log('Failed to delete');
              end;
            end;
          until FindNext(SR) <> 0;
        finally
          SysUtils.FindClose(SR);
        end;
      // Then delete the directory itself
      if ContentDeleted then begin
        //Log('Content deleted, deleting "%s"', [FileName]);
        SetFileAttributes(PChar(FileName), Attr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
        if RemoveDirectory(PChar(FileName)) then begin
          //Log('- Deleted successfully');
          Result := True;
          Break;
        end;
      end;
      // If I got here, then the directory couldn't be deleted
      case AskRetrySkipCancel1(MFailedEraseDir, FileName, SkipAllFailedDeletes) of
        rscSkip:
          Break;
        rscCancel:
          Abort;
      end;
    end;
  until False;
  //Log('Result: %d', [Integer(Result)]);
end;

function TSynchronizeTask.DeleteFileMissingOnTheOtherPanel(const FileName: string; var CanDelete, CanDeleteReadOnly: TSyncFileActionFlag; var SkipAllFailedDeletes: boolean; DeleteOnLeft: boolean): boolean;
var
  Attr: DWORD;
begin
  Result := False;
  ShowProgress(FileName, '', 0, 1, MDeletingFile, MNoLngStringDefined);
  //Log('Deleting file "%s"', [FileName]);
  repeat
    if not GetFileAttributes(FileName, Attr) then begin
      //Log('- File does not exist');
      Result := True;
      Break;
    end
    else begin
      if not AskDeleteConfirmation(FileName, CanDelete, CanDeleteReadOnly, DeleteOnLeft) then begin
        //Log('- User did not allow deleting the destination file');
        Break;
      end;
      SetFileAttributes(PChar(FileName), Attr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
      if DeleteFile(PChar(FileName)) then begin
        //Log('- Deleted successfully');
        Result := True;
        Break;
      end
      else begin
        case AskRetrySkipCancel1(MFailedEraseFile, FileName, SkipAllFailedDeletes) of
          rscSkip:
            Break;
          rscCancel:
            Abort;
        end;
      end;
    end;
  until False;
  //Log('- Result: %d', [Integer(Result)]);
end;

destructor TSynchronizeTask.Destroy;
begin
  FreeAndNil(fProgress);
  if FileName <> '' then
    if FileExists(FileName) then
      SysUtils.DeleteFile(FileName);
  inherited;
end;

procedure TSynchronizeTask.Edit(const FileName: string);
begin
  FarApi.Editor(PFarChar(FileName), GetMsg(MCmpTitle), 0, 0, -1, -1,
    EF_NONMODAL or EF_IMMEDIATERETURN,
    0, 1, {$IFDEF FAR3} CP_DEFAULT {$ELSE} CP_AUTODETECT {$ENDIF}
  );
end;

procedure TSynchronizeTask.EscapePressedEvent(Sender: TObject; var Value: boolean);
begin
  Value := FarMessage([MEscTitle, MEscBody, MOk, MCancel], FMSG_WARNING, 2) = 0;
end;

function TSynchronizeTask.GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
var
  LastError: DWORD;
  SR: TSearchRec;
begin
  Attributes := Windows.GetFileAttributes(PChar(FileName));
  if Attributes <> INVALID_FILE_ATTRIBUTES then
    Result := True
  else begin
    LastError := GetLastError;
    if (LastError = ERROR_FILE_NOT_FOUND) or (LastError = ERROR_PATH_NOT_FOUND) or (LastError = ERROR_INVALID_NAME) then
      Result := False
    else
      if FindFirst(FileName, faAnyFile, SR) = 0 then
        try
          Attributes := SR.Attr;
          Result := True;
        finally
          SysUtils.FindClose(SR);
        end
      else
        Result := False;
  end;
end;

function CopyProgressRoutine(TotalFileSize, TotalBytesTransferred, StreamSize, StreamBytesTransferred: int64; dwStreamNumber, dwCallbackReason: DWORD; hSourceFile, hDestinationFile: THandle; lpData: TSynchronizeTask): DWORD; stdcall;
begin
  Result := PROGRESS_CONTINUE;
  if lpData <> nil then
    try
      lpData.ShowFileCopyProgress(TotalBytesTransferred, TotalFileSize);
    except
      on EAbort do
        begin
          lpData.fCopyFileAborted := True;
          Result := PROGRESS_CANCEL;
        end;
      on Exception do
        Result := PROGRESS_CONTINUE;
    end;
end;

function TSynchronizeTask.InternalCopyFileWithProgress(const SourceFileName, DestinationFileName: string; out ErrorCode: DWORD): boolean;
var SrcAttr, DestAttr: DWORD;
begin
  if GetFileAttributes(SourceFileName, SrcAttr) then begin
    if GetFileAttributes(DestinationFileName, DestAttr) then begin
      SetFileAttributes(PChar(DestinationFileName), DestAttr and (not (FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or FILE_ATTRIBUTE_SYSTEM)));
      DeleteFile(PChar(DestinationFileName));
    end;
    fCopyFileSourceFileName := SourceFileName;
    fCopyFileDestinationFileName := DestinationFileName;
    fCopyFileAborted := False;
    Result := CopyFileEx(PChar(SourceFileName), PChar(DestinationFileName), @CopyProgressRoutine, Self, nil, 0);
    ErrorCode := GetLastError;
    if Result then
      SetFileAttributes(PChar(DestinationFileName), SrcAttr)
    else if fCopyFileAborted then
      Abort;
  end
  else
    Result := False;
end;

function TSynchronizeTask.LoadFiles(out List: TDifferenceList; const FileName: string): boolean;
begin
  Result := False;
  List := TDifferenceList.Create(True);
  try
    List.LoadFromFile(FileName);
    Result := True;
  finally
    if not Result then
      FreeAndNil(List);
  end;
end;

function TSynchronizeTask.RemoveNTSchematicsFromFileName(const FileName: string): string;
const
  LocalResource = '\\?\';
  NetworkResource = '\\?\UNC\';
begin
  if UpperCase(Copy(FileName, 1, Length(NetworkResource))) = NetworkResource then
    Result := '\\' + Copy(FileName, Succ(Length(NetworkResource)), MaxInt)
  else if Copy(FileName, 1, Length(LocalResource)) = LocalResource then
    Result := Copy(FileName, Succ(Length(LocalResource)), MaxInt)
  else
    Result := FileName;
  //Log('RemoveNTSchematicsFromFileName("%s") -> "%s"', [FileName, Result]);
end;

procedure TSynchronizeTask.SetupKeyBar;
{$IFDEF FAR3}
const
  KEY_CTRL_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_CTRL_F3: TFarKey = (VirtualKeyCode: VK_F3; ControlKeyState: LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED);
  KEY_ALT_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F3: TFarKey = (VirtualKeyCode: VK_F3; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
{$ENDIF}
var
  {$IFDEF FAR3}
  SetKeyBarTitles: TFarSetKeyBarTitles;
  KeyBarTitles: TKeyBarTitles;
  KeyBarLabels: packed array[0..5] of TKeyBarLabel;
  {$ELSE}
  KeyBarTitles: TKeyBarTitles;
  {$ENDIF}
begin
  {$IFDEF FAR3}
  KeyBarLabels[0].Key := KEY_CTRL_F1;
  KeyBarLabels[0].Text := GetMsg(MKeyEditLeft);
  KeyBarLabels[0].LongText := GetMsg(MKeyEditLeftLong);
  KeyBarLabels[1].Key := KEY_CTRL_F2;
  KeyBarLabels[1].Text := GetMsg(MKeyEditRight);
  KeyBarLabels[1].LongText := GetMsg(MKeyEditRightLong);
  KeyBarLabels[2].Key := KEY_ALT_F1;
  KeyBarLabels[2].Text := GetMsg(MKeyOverwriteLeft);
  KeyBarLabels[2].LongText := GetMsg(MKeyOverwriteLeftLong);
  KeyBarLabels[3].Key := KEY_ALT_F2;
  KeyBarLabels[3].Text := GetMsg(MKeyOverwriteRight);
  KeyBarLabels[3].LongText := GetMsg(MKeyOverwriteRightLong);
  KeyBarLabels[4].Key := KEY_ALT_F3;
  KeyBarLabels[4].Text := GetMsg(MKeyCompare);
  KeyBarLabels[4].LongText := GetMsg(MKeyCompareLong);
  KeyBarLabels[5].Key := KEY_CTRL_F3;
  KeyBarLabels[5].Text := GetMsg(MKeyFileInfo);
  KeyBarLabels[5].LongText := GetMsg(MKeyFileInfoLong);
  KeyBarTitles.CountLabels := Length(KeyBarLabels);
  KeyBarTitles.Labels := @KeyBarLabels[0];
  SetKeyBarTitles.StructSize := Sizeof(SetKeyBarTitles);
  SetKeyBarTitles.Titles := @KeyBarTitles;
  FarApi.EditorControl(Self.EditorID, ECTL_SETKEYBAR, 0, @SetKeyBarTitles);
  {$ELSE}
  FillChar(KeyBarTitles, Sizeof(KeyBarTitles), 0);
  KeyBarTitles.CtrlTitles[0] := GetMsg(MKeyEditLeft);
  KeyBarTitles.CtrlTitles[1] := GetMsg(MKeyEditRight);
  KeyBarTitles.CtrlTitles[2] := GetMsg(MKeyFileInfo);
  KeyBarTitles.AltTitles[0] := GetMsg(MKeyOverwriteLeft);
  KeyBarTitles.AltTitles[1] := GetMsg(MKeyOverwriteRight);
  KeyBarTitles.AltTitles[2] := GetMsg(MKeyCompareLong);
  FarApi.EditorControl(ECTL_SETKEYBAR, @KeyBarTitles);
  {$ENDIF}
end;

function TSynchronizeTask.ShowEditor: boolean;
begin
  Result := inherited ShowEditor(GetMsg(MCmpTitle), 0, 0, -1, -1,
        EF_NONMODAL or EF_IMMEDIATERETURN or EF_DELETEONLYFILEONCLOSE or EF_DISABLEHISTORY {$IFDEF FAR3} or EF_DISABLESAVEPOS {$ENDIF} ,
        0, 1, {$IFDEF UNICODE} CP_UTF8 {$ELSE} CP_ANSI {$ENDIF} );
end;

procedure TSynchronizeTask.ShowFileCopyProgress(Progress, Max: int64);
begin
  ShowProgress(fCopyFileSourceFileName, fCopyFileDestinationFileName, Progress, Max, MCopying, MCopyingTo);
end;

procedure TSynchronizeTask.ShowProgress(const SourceFileName, DestinationFileName: string; Progress, Max: int64; Msg1, Msg2: TMessages);
begin
  if fProgress.NeedsRefresh then begin
    fProgress.Show(Integer(MCmpTitle), Integer(Msg1), Integer(Msg2), SourceFileName, DestinationFileName, Progress, Max);
  end;
end;

procedure TSynchronizeTask.AfterShowEditor;
begin
  inherited;
  SetupKeyBar;
end;

function TSynchronizeTask.AskDeleteConfirmation(const FileName: string; var CanDelete, CanDeleteReadOnly: TSyncFileActionFlag; const DeleteOnLeft: boolean): boolean;
var
  Attr: DWORD;
  Dlg: TDeleteConfirmationDialog;
begin
  //Log('Asking for delete confirmation.');
  //Log('- Filename = %s', [Filename]);
  if not GetFileAttributes(FileName, Attr) then
    Result := True
  else begin
    //Log('- File exists');
    Result := False;
    Dlg := TDeleteConfirmationDialog.Create;
    try
      Dlg.FileOnLeft := DeleteOnLeft;
      Dlg.ReadOnlyFile := (Attr and FILE_ATTRIBUTE_READONLY) <> 0;
      //Log('- Read only: %d', [Integer(DeleteDlg.ReadOnlyFile)]);
      Dlg.FileName := FileName;
      Dlg.IsDirectory := (Attr and FILE_ATTRIBUTE_DIRECTORY) <> 0;
      //Log('- Directory: %d', [Integer(DeleteDlg.IsDirectory)]);
      //Log('Asking user for overwrite confirmation. ReadOnly=%d, CanOverwrite=%d, CanOverwriteReadOnly=%d', [Integer(OverwriteDlg.ReadOnlyDestination), Integer(CanOverwrite), Integer(CanOverwriteReadOnly)]);
      if Dlg.ReadOnlyFile then begin
        if CanDeleteReadOnly = afAll then
          Result := True
        else if (CanDeleteReadOnly = afAsk) and (not SilentMode) then
          case Dlg.Execute of
            ocbYes:
              Result := True;
            ocbYesToAll:
              begin
                Result := True;
                CanDeleteReadOnly := afAll;
                CanDelete := afAll;
              end;
            ocbNo:
              Result := False;
            ocbNoToAll:
              begin
                Result := False;
                CanDeleteReadOnly := afNone;
              end;
            ocbCancel:
              Abort;
          end;
        end
      else begin
        if CanDelete = afAll then
          Result := True
        else if (CanDelete = afAsk) and (not SilentMode) then
          case Dlg.Execute of
            ocbYes:
              Result := True;
            ocbYesToAll:
              begin
                Result := True;
                CanDelete := afAll;
              end;
            ocbNo:
              Result := False;
            ocbNoToAll:
              begin
                Result := False;
                CanDelete := afNone;
                CanDeleteReadOnly := afNone;
              end;
            ocbCancel:
              Abort;
          end;
       end;
    finally
      FreeAndNil(Dlg);
    end;
  end;
  //Log('- Result = %d', [Integer(Result)]);
end;

function TSynchronizeTask.AskOverwriteConfirmation(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; const FromRightToLeft: boolean): boolean;
var
  DestAttr: DWORD;
  Dlg: TOverwriteConfirmationDialog;
  Again: boolean;
begin
  //Log('AskOverwriteConfirmation("%s", "%s", %d, %d, %d)', [SourceFileName, DestinationFileName, Integer(CanOverwrite), Integer(CanOverwriteReadOnly), Integer(FromRightToLeft)]);
  if not GetFileAttributes(DestinationFileName, DestAttr) then
    Result := True
  else begin
    Result := False;
    repeat
      Again := False;
      Dlg := TOverwriteConfirmationDialog.Create;
      try
        Dlg.FromRightToLeft := FromRightToLeft;
        Dlg.ReadOnlyDestination := (DestAttr and FILE_ATTRIBUTE_READONLY) <> 0;
        Dlg.SourceFileName := SourceFileName;
        Dlg.DestinationFileName := DestinationFileName;
        //Log('Asking user for overwrite confirmation. ReadOnly=%d, CanOverwrite=%d, CanOverwriteReadOnly=%d', [Integer(Dlg.ReadOnlyDestination), Integer(CanOverwrite), Integer(CanOverwriteReadOnly)]);
        if Dlg.ReadOnlyDestination then begin
          if CanOverwriteReadOnly in [afAll, afYes] then
            Result := True
          else if (CanOverwriteReadOnly = afAsk) and (not SilentMode) then
            case Dlg.Execute of
              ocbYes:
                Result := True;
              ocbYesToAll:
                begin
                  Result := True;
                  CanOverwriteReadOnly := afAll;
                  CanOverwrite := afAll;
                end;
              ocbNo:
                Result := False;
              ocbNoToAll:
                begin
                  Result := False;
                  CanOverwriteReadOnly := afNone;
                end;
              ocbCompare:
                begin
                  if FromRightToLeft then
                    VisualCompare(Dlg.DestinationFileName, Dlg.SourceFileName)
                  else
                    VisualCompare(Dlg.SourceFileName, Dlg.DestinationFileName);
                  Again := True;
                end;
              ocbCancel:
                Abort;
            end;
          end
        else begin
          if CanOverwrite in [afAll, afYes] then
            Result := True
          else if (CanOverwrite = afAsk) and (not SilentMode) then
            case Dlg.Execute of
              ocbYes:
                Result := True;
              ocbYesToAll:
                begin
                  Result := True;
                  CanOverwrite := afAll;
                end;
              ocbNo:
                Result := False;
              ocbNoToAll:
                begin
                  Result := False;
                  CanOverwrite := afNone;
                  CanOverwriteReadOnly := afNone;
                end;
              ocbCompare:
                begin
                  if FromRightToLeft then
                    VisualCompare(Dlg.DestinationFileName, Dlg.SourceFileName)
                  else
                    VisualCompare(Dlg.SourceFileName, Dlg.DestinationFileName);
                  Again := True;
                end;
              ocbCancel:
                Abort;
             end;
          end;
        //Log('- Result = %d', [Integer(Result)]);
      finally
        FreeAndNil(Dlg);
      end;
    until not Again;
  end;
end;

function TSynchronizeTask.AskRetrySkipCancel1(Msg: TMessages; const FileName: string; var SkipAll: boolean): TSyncFileRetrySkipCancel;
begin
  {$IFDEF SIMPLE_RETRYSKIPCANCEL}
  Result := rscSkip;
  if not SilentMode then begin
    if not SkipAll then begin
      case FarMessage([MWarning, Msg, MExtraString1, MRetry, MSkip, MSkipAll, MCancel], [PFarChar(FileName)], FMSG_WARNING, 4) of
        0: Result := rscRetry;
        1: Result := rscSkip;
        2: begin
             Result := rscSkip;
             SkipAll := True;
           end;
        3: Result := rscCancel;
      end;
    end;
  end;
  {$ELSE}
  Result := AskRetrySkipCancel2(Msg, MNoLngStringDefined, FileName, '', SkipAll);
  {$ENDIF}
end;

function TSynchronizeTask.AskRetrySkipCancel2(Msg1, Msg2: TMessages; const FileName1, FileName2: string; var SkipAll: boolean): TSyncFileRetrySkipCancel;
{$IFNDEF SIMPLE_RETRYSKIPCANCEL}
var
  Dialog: TRetrySkipCancelDialog;
{$ENDIF}
begin
  Result := rscSkip;
  if not SilentMode then begin
    if not SkipAll then begin
      {$IFDEF SIMPLE_RETRYSKIPCANCEL}
      case FarMessage([MWarning, Msg1, MExtraString1, Msg2, MExtraString2, MRetry, MSkip, MSkipAll, MCancel], [PFarChar(FileName1), PFarChar(FileName2)], FMSG_WARNING, 4) of
        0: Result := rscRetry;
        1: Result := rscSkip;
        2: begin
             Result := rscSkip;
             SkipAll := True;
           end;
        3: Result := rscCancel;
      end;
      {$ELSE}
      Dialog := TRetrySkipCancelDialog.Create;
      try
        Dialog.Message1 := Msg1;
        Dialog.FileName1 := FileName1;
        Dialog.Message2 := Msg2;
        Dialog.FileName2 := FileName2;
        case Dialog.Execute of
          ocbYes, ocbYesToAll:
            Result := rscRetry;
          ocbNo:
            Result := rscSkip;
          ocbNoToAll:
            begin
              Result := rscSkip;
              SkipAll := True;
            end;
          ocbCancel:
            Result := rscCancel;
        end;
      finally
        FreeAndNil(Dialog);
      end;
      {$ENDIF}
    end;
  end;
end;

function TSynchronizeTask.AskWhatToDoWithDifferentFiles(const LeftDirectory, RightDirectory, FileName: string): TSyncFileDirection;
var
  LeftAttr, RightAttr: DWORD;
  Dlg: TWhatToDoWithDifferentFilesDialog;
  Again: boolean;
begin
  //Log('AskWhatToDoWithDifferentFiles("%s", "%s", "%s")', [LeftDirectory, RightDirectory, FileName]);
  if not GetFileAttributes(LeftDirectory + FileName, LeftAttr) then
    Result := sfdSkip
  else if not GetFileAttributes(RightDirectory + FileName, RightAttr) then
    Result := sfdSkip
  else if SilentMode then
    Result := sfdSkip
  else begin
    Result := sfdSkip;
    repeat
      Again := False;
      Dlg := TWhatToDoWithDifferentFilesDialog.Create;
      try
        Dlg.LeftFileName := LeftDirectory + FileName;
        Dlg.RightFileName := RightDirectory + FileName;
        case Dlg.Execute of
          dfbOverwriteLeft:
            Result := sfdCopyToLeft;
          dfbOverwriteRight:
            Result := sfdCopyToRight;
          dfbSkip:
            Result := sfdSkip;
          dfbAlwaysOverwriteLeft:
            Result := sfdCopyToLeftAll;
          dfbAlwaysOverwriteRight:
            Result := sfdCopyToRightAll;
          dfbAlwaysSkip:
            Result := sfdSkipAll;
          dfbCompare:
            begin
              VisualCompare(Dlg.LeftFileName, Dlg.RightFileName);
              Again := True;
            end;
          dfbCancel:
            Abort;
        end;
      finally
        FreeAndNil(Dlg);
      end;
    until not Again;
  end;
  //Log('AskWhatToDoWithDifferentFiles -> %d', [Integer(Result)]);
end;

procedure TSynchronizeTask.BeforeClose(var CanClose: boolean; var Result: TIntPtr);
var
  ReturnToEditor: boolean;
begin
  inherited;
  Result := 0;
  Execute(ReturnToEditor);
  if ReturnToEditor then
    CanClose := False;
end;

function TSynchronizeTask.CopyDirectory(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
var
  SourceAttr, DestAttr: DWORD;
  SourceDir, DestinationDir: string;
  SR: TSearchRec;
begin
  Result := False;
  ShowProgress(SourceFileName, DestinationFileName, 0, 1, MCopying, MCopyingTo);
  //Log('Copying directory "%s" to "%s"', [SourceFileName, DestinationFileName]);
  repeat
    if not GetFileAttributes(SourceFileName, SourceAttr) then begin
      //Log('- Source directory does not exist');
      Break;
    end
    else begin
      if not GetFileAttributes(DestinationFileName, DestAttr) then begin
        ForceDirectories(DestinationFileName);
        if not GetFileAttributes(DestinationFileName, DestAttr) then begin
          //Log('- Cannot create destination directory');
          case AskRetrySkipCancel1(MCantCreateDirectory, DestinationFileName, SkipAllFailedCopies) of
            rscSkip:
              Break;
            rscCancel:
              Abort;
          end;
        end;
      end;
      //Log('Copying files');
      Result := True;
      SourceDir := IncludeTrailingPathDelimiter(SourceFileName);
      DestinationDir := IncludeTrailingPathDelimiter(DestinationFileName);
      if FindFirst(SourceDir + '*.*', faAnyFile, SR) = 0 then
        try
          repeat
            if (SR.Attr and faDirectory) <> 0 then
              if (SR.Name = '.') or (SR.Name = '..') then
                Continue
              else
                CopyDirectory(SourceDir + SR.Name, DestinationDir + SR.Name, CanOverwrite, CanOverwriteReadOnly, SkipAllFailedCopies, FromRightToLeft)
            else
              CopyFile(SourceDir + SR.Name, DestinationDir + SR.Name, CanOverwrite, CanOverwriteReadOnly, SkipAllFailedCopies, FromRightToLeft);
          until FindNext(SR) <> 0;
        finally
          SysUtils.FindClose(SR);
        end;
      Break;
    end;
  until False;
  //Log('- Result: %d', [Integer(Result)]);
end;

function TSynchronizeTask.CopyFile(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
var
  SourceAttr, DestAttr, CopyError: DWORD;
  Handle: THandle;
  Msg: TMessages;
begin
  //Log('CopyFile("%s", "%s", %d, %d, %d)', [SourceFileName, DestinationFileName, Integer(CanOverwrite), Integer(CanOverwriteReadOnly), Integer(FromRightToLeft)]);
  Result := False;
  ShowProgress(SourceFileName, DestinationFileName, 0, 1, MCopying, MCopyingTo);
  repeat
    if not GetFileAttributes(SourceFileName, SourceAttr) then begin
      //Log('- Source file does not exist');
      Break;
    end
    else begin
      if GetFileAttributes(DestinationFileName, DestAttr) then
        if not AskOverwriteConfirmation(SourceFileName, DestinationFileName, CanOverwrite, CanOverwriteReadOnly, FromRightToLeft) then begin
          //Log('- User did not allow overwriting the destination file');
          Break;
        end;
      if InternalCopyFileWithProgress(SourceFileName, DestinationFileName, CopyError) then begin
        //Log('- Copied successfully');
        Result := True;
        Break;
      end
      else begin
        //Log('- Error %d while copying', [CopyError]);
        Msg := MFailedCopySrcFile;
        Handle := CreateFile(PChar(SourceFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
        if Handle = INVALID_HANDLE_VALUE then
          Msg := MFailedOpenSrcFile
        else begin
          CloseHandle(Handle);
          Handle := CreateFile(PChar(DestinationFileName), GENERIC_WRITE, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
          if Handle = INVALID_HANDLE_VALUE then
            if GetLastError = ERROR_FILE_NOT_FOUND then
              Msg := MFailedCreateDstFile
            else
              Msg := MFailedOpenDstFile
          else
            CloseHandle(Handle);
        end;
				SetLastError(CopyError);
        case AskRetrySkipCancel2(Msg, MFailedCopyDstFile, SourceFileName, DestinationFileName, SkipAllFailedCopies) of
          rscSkip:
            Break;
          rscCancel:
            Abort;
        end;
      end;
    end;
  until False;
  //Log('- Result: %d', [Integer(Result)]);
end;

function TSynchronizeTask.Execute(out ReturnToEditor: boolean): boolean;
const
  ConfirmToSyncFileActionFlag: array[boolean] of TSyncFileActionFlag = (afAll, afAsk);
var
  List: TDifferenceList;
  DiffItem: TDifferenceItem;
  Dlg: TSynchronizeDirectoriesDialog;
  DiffType: TDifferenceType;
  i: Integer;
  FN: string;
  OK, SyncAgain: Boolean;
  CanOverwriteLeft, CanOverwriteReadOnlyLeft: TSyncFileActionFlag;
  CanDeleteDirectoriesLeft, CanDeleteReadOnlyDirectoriesLeft: TSyncFileActionFlag;
  CanDeleteFilesLeft, CanDeleteReadOnlyFilesLeft: TSyncFileActionFlag;
  CanOverwriteRight, CanOverwriteReadOnlyRight: TSyncFileActionFlag;
  CanDeleteDirectoriesRight, CanDeleteReadOnlyDirectoriesRight: TSyncFileActionFlag;
  CanDeleteFilesRight, CanDeleteReadOnlyFilesRight: TSyncFileActionFlag;
  SkipAllFailedDeletes, SkipAllFailedCopies: boolean;
  NewTask: TSynchronizeTask;
  DiffFile: TTextStream;
begin
  Result := False;
  ReturnToEditor := True;
  try
    //FarMessage(['DirSync', '', 'Files should be synchronized here:', PChar(FileName), PChar(LeftDirectory), PChar(RightDirectory), '', 'OK'], 0, 1);
    if LoadFiles(List, Self.FileName) then
      try
        if List.Count > 0 then begin
          Dlg := TSynchronizeDirectoriesDialog.Create;
          try
            Dlg.SilentMode := SilentMode;
            for DiffType := Low(TDifferenceType) to High(TDifferenceType) do
              Dlg.DiffTypePresent[DiffType] := List.HasDifference[DiffType];
            case Dlg.Execute of
              sdaCancel:
                begin
                  ReturnToEditor := False;
                end;
              sdaSynchronize:
                begin
                  ReturnToEditor := False;
                  //FarMessage(['DirSync', '', 'Files should be synchronized here:', PChar(IntToStr(List.Count) + ' files total'), '', 'OK'], 0, 1);
                  fProgress := TFarProgress.Create( {$IFDEF FAR3} PLUGIN_GUID, DIALOG_READING_GUID {$ELSE} FarApi.ModuleNumber {$ENDIF} );
                  try
                    //fProgress.GranularityMSec := PROGRESS_GRANULARITY_MSEC;
                    fProgress.CheckForEscape := True;
                    fProgress.OnEscapePressed := EscapePressedEvent;
                    CanOverwriteLeft := ConfirmToSyncFileActionFlag[Dlg.ConfirmOverwriteLeft];
                    CanOverwriteReadOnlyLeft := ConfirmToSyncFileActionFlag[True];
                    CanDeleteDirectoriesLeft := ConfirmToSyncFileActionFlag[True];
                    CanDeleteFilesLeft := ConfirmToSyncFileActionFlag[True];
                    CanDeleteReadOnlyDirectoriesLeft := ConfirmToSyncFileActionFlag[True];
                    CanDeleteReadOnlyFilesLeft := ConfirmToSyncFileActionFlag[True];
                    CanOverwriteRight := ConfirmToSyncFileActionFlag[Dlg.ConfirmOverwriteRight];
                    CanOverwriteReadOnlyRight := ConfirmToSyncFileActionFlag[True];
                    CanDeleteDirectoriesRight := ConfirmToSyncFileActionFlag[True];
                    CanDeleteFilesRight := ConfirmToSyncFileActionFlag[True];
                    CanDeleteReadOnlyDirectoriesRight := ConfirmToSyncFileActionFlag[True];
                    CanDeleteReadOnlyFilesRight := ConfirmToSyncFileActionFlag[True];
                    SkipAllFailedDeletes := False;
                    SkipAllFailedCopies := False;
                    SilentMode := Dlg.SilentMode;
                    try
                      repeat
                        i := 0;
                        while i < List.Count do begin
                          DiffItem := List[i];
                          FN := Trim(DiffItem.FileName);
                          OK := True; // If some difference types should be ignored, consider files with that difference successfully processed
                          case DiffItem.DifferenceType of
                            dtMissingLeft:
                              if Dlg.DeleteRightMissingOnLeft then
                                if DiffItem.IsDirectory then
                                  OK := DeleteDirectoryMissingOnTheOtherPanel(RightDirectory + FN, CanDeleteDirectoriesRight, CanDeleteFilesRight, CanDeleteReadOnlyDirectoriesRight, CanDeleteReadOnlyFilesRight, SkipAllFailedDeletes, False)
                                else
                                  OK := DeleteFileMissingOnTheOtherPanel(RightDirectory + FN, CanDeleteFilesRight, CanDeleteReadOnlyFilesRight, SkipAllFailedDeletes, False)
                              else if Dlg.CopyToLeftPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True)
                                else
                                  OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True);
                            dtMissingRight:
                              if Dlg.DeleteLeftMissingOnRight then
                                if DiffItem.IsDirectory then
                                  OK := DeleteDirectoryMissingOnTheOtherPanel(LeftDirectory + FN, CanDeleteDirectoriesLeft, CanDeleteFilesLeft, CanDeleteReadOnlyDirectoriesLeft, CanDeleteReadOnlyFilesLeft, SkipAllFailedDeletes, True)
                                else
                                  OK := DeleteFileMissingOnTheOtherPanel(LeftDirectory + FN, CanDeleteFilesLeft, CanDeleteReadOnlyFilesLeft, SkipAllFailedDeletes, True)
                              else if Dlg.CopyToRightPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False)
                                else
                                  OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False);
                            dtOlderLeft:
                              if Dlg.CopyToLeftPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True)
                                else
                                  OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True);
                            dtOlderRight:
                              if Dlg.CopyToRightPanel then
                                if DiffItem.IsDirectory then
                                  OK := CopyDirectory(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False)
                                else
                                  OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False);
                            dtDifferent:
                              case Dlg.DifferentFilesAction of
                                dfaAsk:
                                  case AskWhatToDoWithDifferentFiles(LeftDirectory, RightDirectory, FN) of
                                    sfdSkip:
                                      OK := False;
                                    sfdSkipAll:
                                      begin
                                        OK := False;
                                        Dlg.DifferentFilesAction := dfaSkip;
                                      end;
                                    sfdCopyToLeft:
                                      OK := SyncFile(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True);
                                    sfdCopyToLeftAll:
                                      begin
                                        Dlg.DifferentFilesAction := dfaCopyToLeft;
                                        OK := SyncFile(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True);
                                      end;
                                    sfdCopyToRight:
                                      OK := SyncFile(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False);
                                    sfdCopyToRightAll:
                                      begin
                                        Dlg.DifferentFilesAction := dfaCopyToRight;
                                        OK := SyncFile(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False);
                                      end;
                                  end;
                                dfaSkip:
                                  OK := False;
                                dfaCopyToLeft:
                                  if DiffItem.IsDirectory then
                                    OK := CopyDirectory(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True)
                                  else
                                    OK := CopyFile(RightDirectory + FN, LeftDirectory + FN, CanOverwriteLeft, CanOverwriteReadOnlyLeft, SkipAllFailedCopies, True);
                                dfaCopyToRight:
                                  if DiffItem.IsDirectory then
                                    OK := CopyDirectory(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False)
                                  else
                                    OK := CopyFile(LeftDirectory + FN, RightDirectory + FN, CanOverwriteRight, CanOverwriteReadOnlyRight, SkipAllFailedCopies, False);
                              end;
                          end;
                          if OK then
                            List.Delete(i)
                          else
                            Inc(i);
                        end;
                        SyncAgain := False;
                        if SilentMode then begin
                          //Log('Starting non-silent run');
                          SilentMode := False;
                          if List.Count > 0 then
                            SyncAgain := True;
                        end;
                      until not SyncAgain;
                    except
                      on EAbort do
                        ;
                      on Exception do
                        Raise;
                    end;
                    if List.Count > 0 then begin
                      NewTask := TSynchronizeTask.Create(Self.LeftDirectory, Self.RightDirectory);
                      try
                        DiffFile := TTextStream.Create(NewTask.FileName, fmOpenReadWrite or fmShareDenyWrite);
                        try
                          DiffFile.Position := DiffFile.Size;
                          for i := 0 to Pred(List.Count) do begin
                            DiffFile.WriteLine(List[i].AsString);
                          end;
                        finally
                          FreeAndNil(DiffFile);
                        end;
                        if NewTask.ShowEditor then
                          NewTask := nil;
                      finally
                        FreeAndNil(NewTask);
                      end;
                    end;
                  finally
                    FreeAndNil(fProgress);
                  end;
                  Result := True;
                end;
            end;
          finally
            FreeAndNil(Dlg);
          end;
        end;
      finally
        FreeAndNil(List);
      end;
  except
    on E: EAbort do
      ; //Log('Aborted');
    on E: Exception do
      FarMessage([MCmpTitle, MExtraString1, MExtraString2, MOk], [PFarChar(E.ClassName), PChar(E.Message)], FMSG_ERRORTYPE or FMSG_WARNING, 1);
  end;
end;

function TSynchronizeTask.SyncFile(const SourceFileName, DestinationFileName: string; var CanOverwrite, CanOverwriteReadOnly: TSyncFileActionFlag; var SkipAllFailedCopies: boolean; FromRightToLeft: boolean): boolean;
var
  OldCanOverwrite: TSyncFileActionFlag;
begin
  //Log('SyncFile("%s", "%s", %d, %d, %d)', [SourceFileName, DestinationFileName, Integer(CanOverwrite), Integer(CanOverwriteReadOnly), Integer(FromRightToLeft)]);
  OldCanOverwrite := CanOverwrite;
  CanOverwrite := afYes;
  try
    Result := Self.CopyFile(SourceFileName, DestinationFileName, CanOverwrite, CanOverwriteReadOnly, SkipAllFailedCopies, FromRightToLeft);
  finally
    if CanOverwrite = afYes then
      CanOverwrite := OldCanOverwrite;
  end;
end;

procedure TSynchronizeTask.VisualCompare(LeftFileName, RightFileName: string);
var
  Options: TDirSyncOptions;
  Command, Arguments: string;
begin
  //Log('VisualCompare("%s", "%s")', [LeftFileName, RightFileName]);
  Options := TDirSyncOptions.Create;
  try
    Options.Load;
    Command := Options.CompareCommand;
    Arguments := Options.CompareCommandArguments;
    if (Command = '') or (Arguments = '') then
      //TFarUtils.ShowMessage([GetMsg(MWarning), GetMsg(MVisualCompareNotSetUp), GetMsg(MOK)], FMSG_WARNING, 1)
      FarMessage([MWarning, MVisualCompareNotSetUp, MOK], FMSG_WARNING, 1)
    else begin
      //Log('- CompareCommandAcceptsNTFileNames = %d', [Integer(Options.CompareCommandAcceptsNTFileNames)]);
      if not Options.CompareCommandAcceptsNTFileNames then begin
        LeftFileName := RemoveNTSchematicsFromFileName(LeftFileName);
        RightFileName := RemoveNTSchematicsFromFileName(RightFileName);
        //Log('- Translated filenames: "%s", "%s"', [LeftFileName, RightFileName]);
      end;
      Arguments := StringReplace(Arguments, '${LEFTFILE}', AnsiQuotedStr(LeftFileName, '"'), [rfIgnoreCase, rfReplaceAll]);
      Arguments := StringReplace(Arguments, '${RIGHTFILE}', AnsiQuotedStr(RightFileName, '"'), [rfIgnoreCase, rfReplaceAll]);
      RunFile(Command, Arguments);
    end;
  finally
    FreeAndNil(Options);
  end;
end;

end.
